--[[
-- Explosion Class
-- Explosions are very special. They are inside the "Entities" folder, but
-- they are not a subclass of Entity, like the rest. This is because when an
-- explosion is created, it does its thing, and it immediatelly disappears!
--
-- "Its thing" is:
-- * Shakes the camera
-- * Destroying Guardians and destructible Blocks
-- * Pushing the Player, Grenades, Puffs and Debris around
-- * Damaging the player
-- * Spawning some Puffs
--
-- Then the explosion ceases to exist.
--
-- As a result, it doesn't have neither `update` or `draw` methods. All the work
-- is done in the constructor.
--
-- For detecting which items can be destroyed / pushed / damaged by a particular
-- explosion, this class uses bump's world:queryRect several times, with different
-- boxes (the box which destroys is smaller than the box that "pushes stuff around")
--
--]]


local Explosion = class:extend('Explosion')

local width         = 150
local height        = width
local pushRadius    = 50
local maxPushSpeed  = 300
local minPuffs      = 2--15
local maxPuffs      = 7--30

local max = function(a,b)
  return (b>a and b) or a
end
local min = function(a,b)
  return (b<a and b) or a
end

local clamp = function(x, a, b)
  return max(a, min(b, x))
end

local damageFilter = function(other)
  local cname = other.class
  if other.isBackground then return false end
  return cname == 'Player'
      or ( not other.invincible)
end

local pushFilter = function(other)
  local cname = other.class --log("cname"..inspect(cname))
  if cname == "Puff" then return end
  return cname == 'Player' or cname == 'Grenade' or cname == 'Debris' or cname == 'Puff'
end

function Explosion:damageItem(other)
  other:getAttacked(self)
  --[[
  if other.class.name == 'Player' then
    other:takeHit()
  else
    other:destroy()
  end]]
end

function Explosion:pushItem(other)
  local cx, cy = self.x + self.w / 2, self.y + self.h / 2
  local ox, oy = other:getCenter()
  local dx, dy = ox - cx, oy - cy

  dx, dy = clamp(dx, -maxPushSpeed, maxPushSpeed),
           clamp(dy, -maxPushSpeed, maxPushSpeed)
  if     dx > 0 then dx = maxPushSpeed - dx
  elseif dx < 0 then dx = dx - maxPushSpeed
  end
  if     dy > 0 then dy = maxPushSpeed - dy
  elseif dy < 0 then dy = dy - maxPushSpeed
  end
  dx = 5000*other:dir(dx)+dx
  dy = 5000*other:dir(dy)+dy
  if other.vx then
      other.vx = other.vx + dx
      other.vy = other.vy + dy
  end
end

function Explosion:__init__(kwargs)
  --self.super.__init__(kwargs)
  --self.world:remove(self)
  self.map = kwargs.map
  self.world = kwargs.world or (kwargs.map and kwargs.map.world)
  self.lifeTime = kwargs.lifeTime or kwargs.life
  self.w = kwargs.w or kwargs.width or width
  self.h = kwargs.h or kwargs.height or height
  self.x = kwargs.x or 0
  self.y = kwargs.y or 0
  self.source = kwargs.source or "boom.png"
  self.opacity = kwargs.opacity or 1
  self.opacity2 = kwargs.opacity2 or 1
  self.attack= kwargs.attack or 8
  self.passive = kwargs.passive
  self.parent = kwargs.parent
  
  local minPuffs, maxPuffs = kwargs.minPuffs or minPuffs,
      kwargs.maxPuffs or maxPuffs
      
  self.x,self.y = self:ngetCenter()

  --media.sfx.explosion:play()
  getCamera():shake(1,1,1)

  local items, len = self.world:queryRect(self.x,self.y,self.w,self.h, damageFilter) 
  --log("st"..len)
  
  if not self.passive then
    for i=1,len do
        self:damageItem(items[i])
    end
  end 

  local items, len = self.world:queryRect(
    self.x - pushRadius,
    self.y - pushRadius,
    self.w + pushRadius + pushRadius,
    self.h + pushRadius + pushRadius,
    pushFilter
  )

  if not self.passive then
    for i=1,len do
      self:pushItem(items[i])
    end
  end

  for i=1, math.random(minPuffs, maxPuffs) do
    Puff:new( {map = self.map,
              x = math.random(self.x, self.x + self.w),
              y = math.random(self.y, self.y + self.h),
              life = self.lifeTime,
              source = self.source,
              opacity = self.opacity,
              opacity2 = self.opacity2,
              parent = self.parent
              })
  end

end

function Explosion:getCenter()
    return self.x + self.w / 2,
         self.y + self.h / 2
end

function Explosion:ngetCenter()
    return self.x - self.w / 2,
         self.y - self.h / 2
end

return Explosion
