--[[
-- Fire Class
-- Represents a Fire of smoke, created either by explosions or by the Player's propulsion
-- Fires don't interact with anything (they can be displaced by explosions)
-- They gradually change in shape & color until they disappear (lived & lifeTime are used for that)
-- Since Fires continuously change in size, we keep adding and removing them from the world (this is ok,
-- it's the same thing that bump does internally to move things around)
--]]

FIRES = 0
_FIRES = {}
MAX_FIRES = 2
FIRE_OFFSET = 0
MAX_FIRE_GENERATIONS = 0
GENS = 0

local Fire = Entity:extend("Fire")

local defaultVx      = 0
local defaultVy      = 10
local defaultMinSize = 16
local defaultMaxSize = 20

function Fire:__init__(kwargs)
  
  table.insert(_FIRES, self)
  --_FIRES[FIRES] = self
  self._fireID = FIRES+1
  --FIRES = FIRES+1
  Entity.__init__(self,kwargs)

  vx, vy = kwargs.vx or defaultVx, kwargs.vy or defaultVy
  minSize = kwargs.minSize or defaultMinSize
  maxSize = kwargs.maxSize or defaultMaxSize

  self.w = kwargs.w or math.random(minSize, maxSize)
  self.h = kwargs.h or math.random(minSize, maxSize)
  
  self.generation = kwargs.generation or 1
  GENS = (GENS>self.generation and GENS) or self.generation
  --set to 0 if smoke bullet and user
  self.opacity = kwargs.opacity or 255/255
  --set to 0.95 if smoke bullet and user
  self.opacity2 = kwargs.opacity2 or kwargs.outline or 1
  self.source = kwargs.source or "flame.png"
  self.attack = kwargs.attack or 1
  self.spread = kwargs.spread or true
  --if self.generation 
  if GENS > MAX_FIRE_GENERATIONS then
      self.spread = false
  end
  
  if self.opacity>1 then
      self.opacity = self.opacity/255
  end
  
  if self.opacity2>1 then
      self.opacity2 = self.opacity2/255
  end
  
  if kwargs.favor then
      self.opacity = 0
      self.opacity2 = .95
  end
  
  self.life = kwargs.lifeTime or kwargs.life or 10
  
  self.vx, self.vy = vx, vy
  self.dead = false
  
  self.map:addDebris(self)
  self.isBackground = false
  self.collisionType = "cross"
  if FIRES>MAX_FIRES then
      _FIRES[1].dead=true--:destroy()--ie = true
      FIRES = FIRES - 1
  end
end

function Fire:expand(dt)
  local cx,cy = self:getCenter()
  local percent = self.lived / self.lifeTime
  if percent < 0.2 then
    self.w = self.w + (200 + percent) * dt
    self.h = self.h + (200 + percent) * dt
  else
    self.w = self.w + (20 + percent) * dt
  end

  self.x = cx - self.w / 2
  self.y = cy - self.h / 2
end

function Fire:destroy()
--ssert(self.jh) self.jh= 1
log("destroying"..tostring(self).."aka fire "..self._fireID)
    lume.remove(_FIRES,self)--._fireID-FIRE_OFFSET)
    self.super.destroy(self) log(self._fireID..","..FIRE_OFFSET)
    FIRE_OFFSET = FIRE_OFFSET + 1
    assert(lume.find(self.map.debris,self))
    self.map:removeDebris(self)
    assert(not lume.find(self.map.debris,self))
    self.dead = true
end

function Fire:update(dt)
    self._w, self._h = resizeImage(game:getSource(self.source),self.w,self.h)
    self:updateGravity(dt)
    self.vy = 0
    local vx, vy = self.x+self.vx*dt,self.y+self.vy*dt
    
    if 1 then
        local mx, my = self:move(vx,vy,dt)
        self.x, self.y = mx,my
    end
    if not self.onGround then
        self.life = self.life-1*dt
    end
  
    self.life = self.life - dt

    if self.life <= 0 and not self.dead then
        self.dead = true
    else
    end
end

local util = {}

function Fire:getColor()
  local percent = math.min(1, (self.lived / self.lifeTime) * 1.8)

  return 220, -- math.floor(155*percent),
         150,--255 - math.floor(155*percent),
         150--100
end

function Fire:draw()
    local rr,gg,bb,a= love.graphics.getColor()
    love.graphics.setColor(255,255,255,self.opacity or self.opacity2)
    love.graphics.draw(game:getSource(self.source),self.x,self.y,0,self._w,self._h)
    love.graphics.setColor(rr,gg,bb,a)
  
end

function Fire.checkCollision(item, other)
    if other.isBackground then
        return item.spread and "cross"
    end
    if other.isTile and not other.isBackground then
        return "touch"
    end
    
    if other == item.parent then
        return nil
    elseif other.isFire then
        return "touch"--nil
    elseif other.isBullet then
        return "cross"
    elseif other.isCrate then
        return item.collisionType
    elseif other.isTile then
        return "touch"--slide??
    elseif other.isPlayer then
       return "cross"
    elseif not other.isBackground then
        return item.collisionType
    end


end

function Fire:move(x,y,dt)
    if not self.active then
        --return
    end
    self.die = false
    self.life = self.life - dt
    
    if self.life<=0 then
        self.die = true
    end
    
    
    if not self.active then
        return 0,0
    end
    
    local actualX, actualY, collisions, len = self.world:move(self,x,y,
    self.checkCollision)
    local cc
    for u = 1,len do
        cc = collisions[u]
        Entity.resolveCollision(self,cc)
        self:burn(cc.other,dt)
    end
    if self.die then
        self.dead = true--:destroy()
    end
    
    return actualX, actualY
end

function Fire:burn(item, dt)
    if item.tmpBurnt then --error()
        --return
    end
    if not item.life then error(inspect(item,1)) end
    item.life = item.life - self.attack*dt
    item.burnt = true
    item.tmpBurnt = true
    
    self.life = self.life + self.attack*dt
    self.h = self.h + self.attack*dt
    if item.life<=0 then
        Fire:new({
            map = self.map,
            x = item.x,
            y = item.y,
            w = self.w,
            h = self.h,
            parent = self,
            generation = self.generation+1
        })
    end
end

return Fire