--[[
-- Puff Class
-- Represents a Puff of smoke, created either by explosions or by the Player's propulsion
-- Puffs don't interact with anything (they can be displaced by explosions)
-- They gradually change in shape & color until they disappear (lived & lifeTime are used for that)
-- Since puffs continuously change in size, we keep adding and removing them from the world (this is ok,
-- it's the same thing that bump does internally to move things around)
--]]

local Puff = Entity:extend("Puff")
local getUpdateCount = getUpdateCount

--debug.getinfo( 2, 'nS' )

local defaultVx      = 0
local defaultVy      = -10
local defaultMinSize = 2
local defaultMaxSize = 10

function Puff:__init__(kwargs)
  self.super.__init__(self,kwargs)

  vx, vy = kwargs.vx or defaultVx, kwargs.vy or defaultVy
  minSize = kwargs.minSize or defaultMinSize
  maxSize = kwargs.maxSize or defaultMaxSize

  self.w = math.random(minSize, maxSize)
  self.h = math.random(minSize, maxSize)
  
  --set to 0 if smoke bullet and user
  self.opacity = kwargs.opacity or 100/255
  --set to 0.95 if smoke bullet and user
  self.opacity2 = kwargs.opacity2 or kwargs.outline or 1
  self.source = kwargs.source --or "smoke.png"
  if self.source=="smoke.png" then self.opacity2=100/255 end
  
 
  if self.opacity>1 then
      self.opacity = self.opacity/255
  end
  
  if self.opacity2>1 then
      self.opacity2 = self.opacity2/255
  end
  
  if kwargs.favor then
      self.opacity = 0
      self.opacity2 = .95
  end
  
  local l = kwargs.lifeTime or kwargs.life or 0.1
  self.lifeTime = l + math.random()
  self.lived = 0
  self.vx, self.vy = vx, vy
  self.dead = false
  --self.updateCount = self.updateCount+1

  self.isBackground = true
end

function Puff:expand(dt)
  local cx,cy = self:getCenter()
  local percent = self.lived / self.lifeTime
  if percent < 0.2 then
    self.w = self.w + (200 + percent) * dt
    self.h = self.h + (200 + percent) * dt
  else
    self.w = self.w + (20 + percent) * dt
  end

  self.x = cx - self.w / 2
  self.y = cy - self.h / 2
end

function Puff:destroy()
    self.super.destroy(self)
    self.dead = true
end

function Puff:update(dt,c)
    c = c or getUpdateCount()
    
    if c-1<self.updateCount then --set to <= if shadow
        return
    elseif c-1>self.updateCount then
        self.updateCount = c-1
    end
    self.updateCount = self.updateCount + 1
    
  if self.source then
      self._w, self._h = resizeImage(game:getSource(self.source),self.w,self.h)
  end
  
  self.lived = self.lived + dt

  if self.lived >= self.lifeTime  then
    assert(not self.dead)
    self.dead = true
    --self:destroy()
  else
    self:expand(dt)
    local next_l = self.x + self.vx * dt
    local next_t = self.y + self.vy * dt
    self.world:update(self, next_l, next_t)
    self.x, self.y = next_l, next_t
  end
  if self.dead then
      self:destroy()
  end
end

function Puff:getColor()
  local percent = math.min(1, (self.lived / self.lifeTime) * 1.8)
  --return colors.brown[1],colors.brown[2],colors.brown[3]
  return  (math.floor(155*percent))/255,
          (255 - math.floor(155*percent))/255,
          100/255
end

function Puff:draw()
  if not self.source then
        local rr,gg,bb,a= love.graphics.getColor()
        local r,g,b = self:getColor()
        util.drawFilledRectangle(self.x, self.y, self.w, self.h, r,g,b,self.opacity,self.opacity2)
        love.graphics.setColor(rr,gg,bb,a)
        if ty then love.graphics.setColor(colors.brown[1],colors.brown[2],colors.brown[3],25) end
  else
        local rr,gg,bb,a= love.graphics.getColor()
        love.graphics.setColor(255,255,255,self.opacity2)
        local so = game:getSource(self.source)
        local vw,vh = so:getDimensions()
        love.graphics.draw(game:getSource(self.source),self.x+vw/2,self.y+vh/2,0,self._w,self._h,
        vw/2,vh/2)
        love.graphics.setColor(rr,gg,bb,a)
  end
  
end

return Puff



